// This code example is created for educational purpose
// by Thorsten Thormaehlen (contact: www.thormae.de).
// It is distributed without any warranty.

package de.unimarburg.vboupdateapp;

import android.app.Activity;
import android.opengl.GLSurfaceView;
import android.os.Bundle;
import android.os.SystemClock;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;
import javax.microedition.khronos.opengles.GL11;

import android.opengl.GLU;
import android.util.Log;
import android.view.MotionEvent;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.Random;

class Renderer {
  public float t = 0.0f;
  public int mode = 1;
  public int ayimutSegs = 8;
  public int polarSegs = 8;
  public boolean recreateVBONotify = false;
  
  private int[] vertBufID = { 0 };
  private int[] indicesBufID = { 0 };
  private short vertSize = 0;
  private Random gen2= new Random(); 
  private int SIZEOF_FLOAT = Float.SIZE / 8;
  private int SIZEOF_SHORT = Short.SIZE / 8;
  
  public void init(GL10 gl10, EGLConfig config) {
    GL11 gl = (GL11) gl10;
    gl.glEnable(GL11.GL_DEPTH_TEST);

    recreateVBOs(gl10);
  }
  
  private void recreateVBOs(GL10 gl10) {

    GL11 gl = (GL11) gl10;
    
    if(vertBufID[0] !=0) gl.glDeleteBuffers( 1, vertBufID, 0);
    if(indicesBufID[0] !=0) gl.glDeleteBuffers( 1, indicesBufID, 0);

    // generating VBO input data
    int totalSegs = ayimutSegs * polarSegs;
    ByteBuffer bb = ByteBuffer.allocateDirect(totalSegs * 4 * SIZEOF_FLOAT);
    bb.order(ByteOrder.nativeOrder());
    FloatBuffer vertexData = bb.asFloatBuffer();
    
    // OpenGL ES does only support short index VBO, therefore the maximum of 
    // totalSegs is 2^15
    ByteBuffer bb2 = ByteBuffer.allocateDirect(totalSegs * 6 * SIZEOF_SHORT);
    bb2.order(ByteOrder.nativeOrder());
    ShortBuffer indicesData = bb2.asShortBuffer();
    
    double ayimutStep = 2.0 * Math.PI / (double)ayimutSegs;
    double polarStep = Math.PI / (double)polarSegs;
    double r = 1.0;
    vertSize = 0;
    Random generator = new Random();
    for(int m=0; m < ayimutSegs; m++) {
      for(int n=0; n < polarSegs; n++) {
        double phi = ayimutStep*m;
        double theta = polarStep * n;

        // random radius
        r = 1.0 - 0.3 * Math.abs(generator.nextDouble());

        // compute xyz from spherical coordinates
        float x = (float) (r * Math.sin(theta) * Math.cos(phi));
        float y = (float) (r * Math.sin(theta) * Math.sin(phi));
        float z = (float) (r * Math.cos(theta));

        vertexData.put(x);
        vertexData.put(y);
        vertexData.put(z);
        vertexData.put(1.0f);

        short q0 = (short)(vertSize % totalSegs);
        short q1 = (short)((vertSize+ayimutSegs) % totalSegs);
        short q2 = (short)((vertSize+ayimutSegs + 1) % totalSegs);
        short q3 = (short)((vertSize+1) % totalSegs);
        
        // create vertex indices
        // first triangle of quad
        indicesData.put(q0);
        indicesData.put(q1);
        indicesData.put(q2);
        // second triangle of quad
        indicesData.put(q2);
        indicesData.put(q3);
        indicesData.put(q0);
        
        vertSize++;
      }
    }

    vertexData.flip();
    indicesData.flip();
    
    // generating vertex VBO
    gl.glGenBuffers(1, vertBufID, 0);
    gl.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertBufID[0]);
    gl.glBufferData(GL11.GL_ARRAY_BUFFER, vertexData.capacity()*SIZEOF_FLOAT, vertexData, GL11.GL_STATIC_DRAW);

    // generating index VBO
    gl.glGenBuffers(1, indicesBufID, 0);
    gl.glBindBuffer(GL11.GL_ELEMENT_ARRAY_BUFFER, indicesBufID[0]);
    gl.glBufferData(GL11.GL_ELEMENT_ARRAY_BUFFER, indicesData.capacity()*SIZEOF_SHORT, indicesData, GL11.GL_STATIC_DRAW);
    
    recreateVBONotify = false;
  }
  
  private void updateRandomVBOVertex(GL10 gl10) {

    GL11 gl = (GL11) gl10;

    double ayimutStep = 2.0 * Math.PI / (double) ayimutSegs;
    double polarStep = Math.PI / (double) polarSegs;
    double r = 1.0;

    // randomly select point
    int m = (int) ((ayimutSegs - 1) * gen2.nextDouble());
    int n = (int) ((polarSegs - 1) * gen2.nextDouble());

    double phi = ayimutStep * m;
    double theta = polarStep * n;

    // random radius
    r = 1.0 - 0.3 * gen2.nextDouble();

    // create xyz from spherical coordinates
    float x = (float) (r * Math.sin(theta) * Math.cos(phi));
    float y = (float) (r * Math.sin(theta) * Math.sin(phi));
    float z = (float) (r * Math.cos(theta));

    int vertPos = (m * polarSegs + n) * 4;

    ByteBuffer bb = ByteBuffer.allocateDirect(4 * SIZEOF_FLOAT);
    bb.order(ByteOrder.nativeOrder());
    FloatBuffer newVertexData = bb.asFloatBuffer();
    newVertexData.put(x);
    newVertexData.put(y);
    newVertexData.put(z);
    newVertexData.put(1.0f);
    newVertexData.flip();

    // update vertex in VBO
    gl.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertBufID[0]);
    gl.glBufferSubData(GL11.GL_ARRAY_BUFFER, vertPos * SIZEOF_FLOAT,
        4 * SIZEOF_FLOAT, newVertexData);
  }
  
  public void resize(GL10 gl10, int width, int height) {
    GL11 gl = (GL11) gl10;
    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL11.GL_PROJECTION);
    gl.glLoadIdentity();
    GLU.gluPerspective(gl, 30.0f, (float) width / (float) height, 0.1f, 50.0f);
  }

  public void display(GL10 gl10) {
    GL11 gl = (GL11) gl10;
    
    if(recreateVBONotify) recreateVBOs(gl);
    
 // update up to 100 random vertices
    for(int i = 0; i < 100 && i  < (ayimutSegs/2); i++) {
      updateRandomVBOVertex(gl);
    }
    
    gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    gl.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_DEPTH_BUFFER_BIT);

    gl.glMatrixMode(GL11.GL_MODELVIEW);
    gl.glLoadIdentity();
    // set camera
    GLU.gluLookAt(gl, 3.5f, -1.0f, 3.5f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f);
    // draw scene
    gl.glRotatef(t, 0.0f, 0.0f, 1.0f);

    // activating VBO
    gl.glBindBuffer(GL11.GL_ARRAY_BUFFER, vertBufID[0]);
    int stride = 0;
    gl.glVertexPointer(4, GL11.GL_FLOAT, stride, 0);
    gl.glEnableClientState(GL11.GL_VERTEX_ARRAY);

    if(mode == 0) {
      gl.glColor4f(1.0f,1.0f,1.0f,1.0f);
    }else{
      gl.glColorPointer(4, GL11.GL_FLOAT, stride, 0);
      gl.glEnableClientState(GL11.GL_COLOR_ARRAY);
    }

    gl.glBindBuffer(GL11.GL_ELEMENT_ARRAY_BUFFER, indicesBufID[0]);
    gl.glDrawElements(GL11.GL_TRIANGLES, vertSize*6, GL11.GL_UNSIGNED_SHORT, 0);

    gl.glDisableClientState(GL11.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL11.GL_COLOR_ARRAY);

    gl.glFlush();
  }

}

class GLSurfaceViewRenderer implements GLSurfaceView.Renderer {

  private Renderer renderer = new Renderer();
  
  final int targetFPS = 60;
  final int targetMillis = 1000 / targetFPS;
  private long startTime = 0;
  
  @Override
  public void onDrawFrame(GL10 gl10) {
    
    float offset = 0.5f;
    renderer.t += offset;
    renderer.display(gl10);
    
    // this code attempts to achieve a target frames-per-second rate
    long timeTillNextRender = startTime + targetMillis;
    long sleepTime = timeTillNextRender - SystemClock.uptimeMillis();
    if (sleepTime > 0) {
      try {
        Thread.sleep(sleepTime);
      } catch (InterruptedException e) {
        e.printStackTrace();
      }
    } else {
      // rendering is too slow
    }
    startTime = SystemClock.uptimeMillis();
  }

  @Override
  public void onSurfaceChanged(GL10 gl10, int width, int height) {
    renderer.resize(gl10, width, height);
  }

  @Override
  public void onSurfaceCreated(GL10 gl10, EGLConfig config) {
    renderer.init(gl10, config);
    startTime = SystemClock.uptimeMillis();
  }

  public Renderer getRenderer() {
    return renderer;
  }
}

public class VboUpdate extends Activity {

  private GLSurfaceView myGLSurfaceView;
  private GLSurfaceViewRenderer myGLSurfaceViewRenderer;
  @Override
  public void onCreate(Bundle savedInstanceState) {
    super.onCreate(savedInstanceState);
    myGLSurfaceView = new GLSurfaceView(this);
    myGLSurfaceViewRenderer = new GLSurfaceViewRenderer();
    myGLSurfaceView.setRenderer(myGLSurfaceViewRenderer);
    // myGLSurfaceView.setRenderMode(GLSurfaceView.RENDERMODE_WHEN_DIRTY);
    setContentView(myGLSurfaceView);
  }

  @Override
  protected void onPause() {
    // Large objects in graphics memory should be de-allocated here
    // This is skipped here for simplicity of the example.
    super.onPause();
    myGLSurfaceView.onPause();
  }

  public boolean onTouchEvent(MotionEvent event) {
    if(event.getAction() == MotionEvent.ACTION_DOWN) {
      if(myGLSurfaceViewRenderer.getRenderer().ayimutSegs < 128) {
        myGLSurfaceViewRenderer.getRenderer().ayimutSegs *= 4;
        myGLSurfaceViewRenderer.getRenderer().polarSegs *= 4;
      }else{
        myGLSurfaceViewRenderer.getRenderer().ayimutSegs = 8;
        myGLSurfaceViewRenderer.getRenderer().polarSegs = 8;
      }
      myGLSurfaceViewRenderer.getRenderer().recreateVBONotify = true;
      return true; 
    }
    return false;
  }
  
  @Override
  protected void onResume() {
    super.onResume();
    // Large objects in graphics memory should be re-created here
    // if they were de-allocated on onPause().
    // This is skipped here for simplicity of the example.
    myGLSurfaceView.onResume();
  }
}
